@@ -140,11 +140,11 @@
 #endif
 
 extern "C" {
-    POLYEXTERNALSYMBOL POLYUNSIGNED PolyChDir(FirstArgument threadId, PolyWord arg);
-    POLYEXTERNALSYMBOL POLYUNSIGNED PolyBasicIOGeneral(FirstArgument threadId, PolyWord code, PolyWord strm, PolyWord arg);
-    POLYEXTERNALSYMBOL POLYUNSIGNED PolyPollIODescriptors(FirstArgument threadId, PolyWord streamVec, PolyWord bitVec, PolyWord maxMillisecs);
-    POLYEXTERNALSYMBOL POLYUNSIGNED PolyTestForInput(FirstArgument threadId, PolyWord strm, PolyWord waitMillisecs);
-    POLYEXTERNALSYMBOL POLYUNSIGNED PolyTestForOutput(FirstArgument threadId, PolyWord strm, PolyWord waitMillisecs);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyChDir(POLYUNSIGNED threadId, POLYUNSIGNED arg);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyBasicIOGeneral(POLYUNSIGNED threadId, POLYUNSIGNED code, POLYUNSIGNED strm, POLYUNSIGNED arg);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyPollIODescriptors(POLYUNSIGNED threadId, POLYUNSIGNED streamVec, POLYUNSIGNED bitVec, POLYUNSIGNED maxMillisecs);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyTestForInput(POLYUNSIGNED threadId, POLYUNSIGNED strm, POLYUNSIGNED waitMillisecs);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyTestForOutput(POLYUNSIGNED threadId, POLYUNSIGNED strm, POLYUNSIGNED waitMillisecs);
 }
 
 // References to the standard streams.  They are only needed if we are compiling
@@ -670,18 +670,18 @@ Handle pollTest(TaskData *taskData, Handle stream)
 // and a time to wait and returns a vector of results.
 // Windows: This is messy because "select" only works on sockets.
 // Do the best we can.
-POLYEXTERNALSYMBOL POLYUNSIGNED PolyPollIODescriptors(FirstArgument threadId, PolyWord streamVector, PolyWord bitVector, PolyWord maxMillisecs)
+POLYEXTERNALSYMBOL POLYUNSIGNED PolyPollIODescriptors(POLYUNSIGNED threadId, POLYUNSIGNED streamVector, POLYUNSIGNED bitVector, POLYUNSIGNED maxMillisecs)
 {
     TaskData *taskData = TaskData::FindTaskForId(threadId);
     ASSERT(taskData != 0);
     taskData->PreRTSCall();
     Handle reset = taskData->saveVec.mark();
-    unsigned maxMilliseconds = (unsigned)maxMillisecs.UnTaggedUnsigned();
+    unsigned maxMilliseconds = (unsigned)PolyWord::FromUnsigned(maxMillisecs).UnTaggedUnsigned();
     Handle result = 0;
 
     try {
-        PolyObject  *strmVec = streamVector.AsObjPtr();
-        PolyObject  *bitVec = bitVector.AsObjPtr();
+        PolyObject  *strmVec = PolyWord::FromUnsigned(streamVector).AsObjPtr();
+        PolyObject  *bitVec = PolyWord::FromUnsigned(bitVector).AsObjPtr();
         POLYUNSIGNED nDesc = strmVec->Length();
         ASSERT(nDesc == bitVec->Length());
         // We should check for interrupts even if we're not going to block.
@@ -838,7 +838,7 @@ static Handle change_dirc(TaskData *taskData, Handle name)
 }
 
 // External call
-POLYUNSIGNED PolyChDir(FirstArgument threadId, PolyWord arg)
+POLYUNSIGNED PolyChDir(POLYUNSIGNED threadId, POLYUNSIGNED arg)
 {
     TaskData *taskData = TaskData::FindTaskForId(threadId);
     ASSERT(taskData != 0);
@@ -902,24 +902,30 @@ Handle modTime(TaskData *taskData, Handle filename)
 {
     TempString cFileName(filename->Word());
     if (cFileName == 0) raise_syscall(taskData, "Insufficient memory", NOMEMORY);
-    /* There are two ways to get this information.
-        We can either use GetFileTime if we are able
-        to open the file for reading but if it is locked
-        we won't be able to.  FindFirstFile is the other
-        alternative.  We have to check that the file name
-        does not contain '*' or '?' otherwise it will try
-        to "glob" this, which isn't what we want here. */
-    WIN32_FIND_DATA wFind;
-    HANDLE hFind;
-    const TCHAR *p;
-    for(p = cFileName; *p; p++)
-        if (*p == '*' || *p == '?')
-            raise_syscall(taskData, "Invalid filename", STREAMCLOSED);
-    hFind = FindFirstFile(cFileName, &wFind);
-    if (hFind == INVALID_HANDLE_VALUE)
-        raise_syscall(taskData, "FindFirstFile failed", GetLastError());
-    FindClose(hFind);
-    return Make_arb_from_Filetime(taskData, wFind.ftLastWriteTime);
+    /* There are several ways to get this information.  The best is to
+    *  open the file and use GetFileTime.  Other possibilities are
+    *  GetFileAttributeEx and FindFirstFile.  If the path is a
+    *  symlink GetFileAttributeEx returns the times for the
+    *  symlink itself.  The definition of OS.FileSys.modTime
+    *  doesn't actually say which we should return but the Unix
+    *  version uses stat rather than lstat.
+    */
+    DWORD dwRes = GetFileAttributes(cFileName);
+    if (dwRes == 0xFFFFFFFF)
+        raise_syscall(taskData, "GetFileAttributes failed", GetLastError());
+    HANDLE hFile = CreateFile(cFileName, GENERIC_READ, FILE_SHARE_READ,
+        NULL, OPEN_EXISTING, dwRes & FILE_ATTRIBUTE_DIRECTORY ? FILE_FLAG_BACKUP_SEMANTICS : 0, NULL);
+    if (hFile == INVALID_HANDLE_VALUE)
+        raise_syscall(taskData, "CreateFile failed", GetLastError());
+    FILETIME lastWrite;
+    if (! GetFileTime(hFile, NULL, NULL, &lastWrite))
+    {
+        DWORD dwErr = GetLastError();
+        CloseHandle(hFile);
+        raise_syscall(taskData, "GetFileTime failed", dwErr);
+    }
+    CloseHandle(hFile);
+    return Make_arb_from_Filetime(taskData, lastWrite);
 }
 
 /* Get file size. */
@@ -928,17 +934,22 @@ Handle fileSize(TaskData *taskData, Handle filename)
     TempString cFileName(filename->Word());
     if (cFileName == 0) raise_syscall(taskData, "Insufficient memory", NOMEMORY);
     /* Similar to modTime*/
-    WIN32_FIND_DATA wFind;
-    HANDLE hFind;
-    const TCHAR *p;
-    for(p = cFileName; *p; p++)
-        if (*p == '*' || *p == '?')
-            raise_syscall(taskData, "Invalid filename", STREAMCLOSED);
-    hFind = FindFirstFile(cFileName, &wFind);
-    if (hFind == INVALID_HANDLE_VALUE)
-        raise_syscall(taskData, "FindFirstFile failed", GetLastError());
-    FindClose(hFind);
-    return Make_arb_from_32bit_pair(taskData, wFind.nFileSizeHigh, wFind.nFileSizeLow);
+    DWORD dwRes = GetFileAttributes(cFileName);
+    if (dwRes == 0xFFFFFFFF)
+        raise_syscall(taskData, "GetFileAttributes failed", GetLastError());
+    HANDLE hFile = CreateFile(cFileName, GENERIC_READ, FILE_SHARE_READ,
+        NULL, OPEN_EXISTING, dwRes & FILE_ATTRIBUTE_DIRECTORY ? FILE_FLAG_BACKUP_SEMANTICS : 0, NULL);
+    if (hFile == INVALID_HANDLE_VALUE)
+        raise_syscall(taskData, "CreateFile failed", GetLastError());
+    LARGE_INTEGER fileSize;
+    if (!GetFileSizeEx(hFile, &fileSize))
+    {
+        DWORD dwErr = GetLastError();
+        CloseHandle(hFile);
+        raise_syscall(taskData, "GetFileSizeEx failed", dwErr);
+    }
+    CloseHandle(hFile);
+    return Make_arb_from_32bit_pair(taskData, fileSize.HighPart, fileSize.LowPart);
 }
 
 /* Set file modification and access times. */
@@ -951,16 +962,19 @@ Handle setTime(TaskData *taskData, Handle fileName, Handle fileTime)
     FILETIME ft;
     /* Get the file time. */
     getFileTimeFromArb(taskData, fileTime, &ft);
-    /* Open an existing file with write access. We need that
-        for SetFileTime. */
+    DWORD dwRes = GetFileAttributes(cFileName);
+    if (dwRes == 0xFFFFFFFF)
+        raise_syscall(taskData, "GetFileAttributes failed", GetLastError());
+    // Open an existing file with write access. We need that for SetFileTime.
+    // Directories require FILE_FLAG_BACKUP_SEMANTICS.
     HANDLE hFile = CreateFile(cFileName, GENERIC_WRITE, 0, NULL, OPEN_EXISTING,
-                FILE_ATTRIBUTE_NORMAL, NULL);
+        dwRes & FILE_ATTRIBUTE_DIRECTORY ? FILE_FLAG_BACKUP_SEMANTICS : FILE_ATTRIBUTE_NORMAL, NULL);
     if (hFile == INVALID_HANDLE_VALUE)
         raise_syscall(taskData, "CreateFile failed", GetLastError());
     /* Set the file time. */
     if (!SetFileTime(hFile, NULL, &ft, &ft))
     {
-        int nErr = GetLastError();
+        DWORD nErr = GetLastError();
         CloseHandle(hFile);
         raise_syscall(taskData, "SetFileTime failed", nErr);
     }
@@ -1332,7 +1346,7 @@ static Handle IO_dispatch_c(TaskData *taskData, Handle args, Handle strm, Handle
 
 // General interface to IO.  Ideally the various cases will be made into
 // separate functions.
-POLYUNSIGNED PolyBasicIOGeneral(FirstArgument threadId, PolyWord code, PolyWord strm, PolyWord arg)
+POLYUNSIGNED PolyBasicIOGeneral(POLYUNSIGNED threadId, POLYUNSIGNED code, POLYUNSIGNED strm, POLYUNSIGNED arg)
 {
     TaskData *taskData = TaskData::FindTaskForId(threadId);
     ASSERT(taskData != 0);
@@ -1357,7 +1371,7 @@ POLYUNSIGNED PolyBasicIOGeneral(FirstArgument threadId, PolyWord code, PolyWord
     else return result->Word().AsUnsigned();
 }
 
-POLYEXTERNALSYMBOL POLYUNSIGNED PolyTestForInput(FirstArgument threadId, PolyWord strm, PolyWord waitMillisecs)
+POLYEXTERNALSYMBOL POLYUNSIGNED PolyTestForInput(POLYUNSIGNED threadId, POLYUNSIGNED strm, POLYUNSIGNED waitMillisecs)
 {
     TaskData *taskData = TaskData::FindTaskForId(threadId);
     ASSERT(taskData != 0);
@@ -1366,9 +1380,9 @@ POLYEXTERNALSYMBOL POLYUNSIGNED PolyTestForInput(FirstArgument threadId, PolyWor
     bool result = false;
 
     try {
-        WinStream *stream = *(WinStream **)(strm.AsObjPtr());
+        WinStream *stream = *(WinStream **)(PolyWord::FromUnsigned(strm).AsObjPtr());
         if (stream == 0) raise_syscall(taskData, "Stream is closed", STREAMCLOSED);
-        result = stream->testForInput(taskData, (unsigned)waitMillisecs.UnTaggedUnsigned());
+        result = stream->testForInput(taskData, (unsigned)PolyWord::FromUnsigned(waitMillisecs).UnTaggedUnsigned());
     }
     catch (KillException &) {
         processes->ThreadExit(taskData); // TestAnyEvents may test for kill
@@ -1380,7 +1394,7 @@ POLYEXTERNALSYMBOL POLYUNSIGNED PolyTestForInput(FirstArgument threadId, PolyWor
     return TAGGED(result? 1: 0).AsUnsigned();
 }
 
-POLYEXTERNALSYMBOL POLYUNSIGNED PolyTestForOutput(FirstArgument threadId, PolyWord strm, PolyWord waitMillisecs)
+POLYEXTERNALSYMBOL POLYUNSIGNED PolyTestForOutput(POLYUNSIGNED threadId, POLYUNSIGNED strm, POLYUNSIGNED waitMillisecs)
 {
     TaskData *taskData = TaskData::FindTaskForId(threadId);
     ASSERT(taskData != 0);
@@ -1389,9 +1403,9 @@ POLYEXTERNALSYMBOL POLYUNSIGNED PolyTestForOutput(FirstArgument threadId, PolyWo
     bool result = false;
 
     try {
-        WinStream *stream = *(WinStream **)(strm.AsObjPtr());
+        WinStream *stream = *(WinStream **)(PolyWord::FromUnsigned(strm).AsObjPtr());
         if (stream == 0) raise_syscall(taskData, "Stream is closed", STREAMCLOSED);
-        result = stream->testForOutput(taskData, (unsigned)waitMillisecs.UnTaggedUnsigned());
+        result = stream->testForOutput(taskData, (unsigned)PolyWord::FromUnsigned(waitMillisecs).UnTaggedUnsigned());
     }
     catch (KillException &) {
         processes->ThreadExit(taskData); // TestAnyEvents may test for kill
